/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozCmgui.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "mozCmgui.h"
#include <nsCOMPtr.h>
#include <nsIDirectoryService.h>
#include <stdio.h>
#include <nsStringAPI.h>
#include <omniORB4/CORBA.h>
#include "Icorba4xpcom.h"
#include <nspr/nspr.h>
#include <nsIRunnable.h>
#ifdef XP_WIN32
#include <windows.h>
#include <nsIThread.h>
#include <assert.h>
#endif

#include "nsIStreamListener.h"
#include "nsIURI.h"
#include "necko/nsNetUtil.h"
//#include "nsEmbedString.h" // should be using embedstring api
//                           // but just use this hack for now until
//                           // figure out how to do this correctly
typedef nsString nsEmbedString;
typedef nsCString nsEmbedCString;

#include "mozDownloadMonitor.h"

NS_IMPL_ISUPPORTS1(mozCmgui, mozICmgui)

mozCmgui::mozCmgui()
{
#ifdef XP_WIN
  corbaThread = new mozCmguiCORBAThread();
#endif
}

mozCmgui::~mozCmgui()
{
#ifdef XP_WIN
  corbaThread->AutoDestruct();
#endif
}

NS_IMETHODIMP
mozCmgui::LaunchCmgui(CmguiIApplicationBootstrap *appb)
{
  // Find the components directory...
  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  nsCOMPtr<nsIFile> comdir;
  nsresult rv =
    dsp->Get("ComsD", NS_GET_IID(nsIFile), getter_AddRefs(comdir));
  if (rv != NS_OK)
  {
    printf("Can't find ComsD.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }
#ifdef XP_WIN
  rv = comdir->Append(NS_LITERAL_STRING("ccbridge.exe"));
#else
  rv = comdir->Append(NS_LITERAL_STRING("ccbridge"));
#endif
  
  PRBool exists;
  comdir->Exists(&exists);
  if (!exists)
  {
    nsresult rv =
      dsp->Get("ProfD", NS_GET_IID(nsIFile), getter_AddRefs(comdir));
    if (rv != NS_OK)
    {
      printf("Can't find ProfD.\n");
      fflush(stdout);
      return NS_ERROR_FAILURE;
    }
    comdir->Append(NS_LITERAL_STRING("extensions"));
    comdir->Append(NS_LITERAL_STRING("{c3405567-4d87-48a0-9c6b-dcccad16d59c}"));
    comdir->Append(NS_LITERAL_STRING("components"));
#ifdef XP_WIN
    rv = comdir->Append(NS_LITERAL_STRING("ccbridge.exe"));
#else
    rv = comdir->Append(NS_LITERAL_STRING("ccbridge"));
#endif
  }

  nsCAutoString bridgeCPath;
  rv = comdir->GetNativePath(bridgeCPath);
  if (rv != NS_OK)
  {
    printf("Can't get native path.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }

#ifdef DEBUG_MODE
  printf("Creating ORB instance...\n");
  fflush(stdout);
#endif
  // Convert appb to a string...
  nsCOMPtr<CORBA_IORBProxy> orbp =
    do_CreateInstance("@cellml.org/CORBA/ORB;1");
  if (orbp == nsnull)
  {
    printf("Can't get an ORB.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }

  PRBool success;
  if (orbp->Load_module(
                        "{c3405567-4d87-48a0-9c6b-dcccad16d59c}",
                        "mozcmgui",
                        &success) != NS_OK)
  {
    printf("Warning: Load_module for mozcmgui failed.\n");
    fflush(stdout);
  }

  char* objstr;
#ifdef DEBUG_MODE
  printf("Stringfying an object...\n");
  fflush(stdout);
#endif
  rv = orbp->Object_to_string(NS_STATIC_CAST(nsISupports*,appb), &objstr);
  if (rv != NS_OK)
  {
    printf("Object_to_string failed.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }

#ifdef DEBUG_MODE
  printf("Getting bridgeCPath...\n");
  fflush(stdout);
#endif
  const char* bcp = bridgeCPath.get();
  const char* bridge_argv[] =
    {
#ifdef VALGRIND_MODE
      "valgrind",
#endif
#ifdef DEBUG_MODE
      "ddd", "--args",
#endif
      bcp,
      objstr,
      NULL
    };

#ifdef DEBUG_MODE
  printf("Creating process...\n");
  fflush(stdout);
  char buf[1024];
  sprintf(buf, "c:/cygwin32/bin/gdb.exe --args \"%s\" \"%s\"",
           bcp, objstr);
#endif

#ifdef DEBUG_MODE_WIN32
  PROCESS_INFORMATION pri;
  STARTUPINFO si;
  memset(&si, 0, sizeof(si));
  si.cb = sizeof(si);

  CreateProcess
    (
     "c:/cygwin/bin/gdb.exe", buf, NULL, NULL, FALSE,
     CREATE_NEW_CONSOLE, NULL, NULL, &si, &pri
    );
#else
  PRProcess* bridgeProcess =
    PR_CreateProcess(
#ifdef DEBUG_MODE
                     "c:/cygwin/bin/gdb.exe",
#elif VALGRIND_MODE
                     "/home/amil082/usr/bin/valgrind",
#else
                     bcp /* "/usr/bin/ddd" */,
#endif
                     (char**)bridge_argv, NULL, NULL);
#ifdef DEBUG_MODE
  printf("Freeing string argument..\n");
  fflush(stdout);
#endif
  nsMemory::Free(objstr);
  if (bridgeProcess == nsnull)
  {
    printf("Can't fork bridge process.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }
#ifdef DEBUG_MODE
  printf("Detaching the process...\n");
  fflush(stdout);
#endif
  PR_DetachProcess(bridgeProcess);
#endif

  return NS_OK;
}



NS_IMETHODIMP
mozCmgui::DefineMemoryBlockFromUrl(CmguiICommandData* cdata,
				   mozIDownloadMonitor* dm,
				   const nsACString& file,
                                   const nsACString& memoryName)
{
  nsresult rv;
  nsCOMPtr<nsIURI> uri;

  rv = NS_NewURI(getter_AddRefs(uri), file, nsnull, nsnull, nsnull);
  if ( NS_SUCCEEDED( rv ) ) {
    //printf( "...NS_NewURI completed OK\n" );
  } else {
    printf( "%s %d: NS_NewURI failed, rv=0x%08X\n", (char*)__FILE__, (int)__LINE__, (int)rv );
    return rv;
  }

  if (dm != nsnull)
    dm->IncrementCounter();

  DnloadListener *listener = new DnloadListener(memoryName, dm, cdata);

  if (listener == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;
  NS_ADDREF(listener);

  rv = NS_OpenURI(listener, nsnull , uri, nsnull, nsnull, nsnull);

  NS_RELEASE(listener);

  if ( NS_SUCCEEDED( rv ) ) {
    //printf( "...NS_OpenURI completed OK\n" );
  } else {
    printf( "%s %d: NS_OpenURI failed, rv=0x%08X\n", (char*)__FILE__, (int)__LINE__, (int)rv );
    return rv;
  }

  return NS_OK;
}



DnloadListener::DnloadListener(const nsACString& aMemoryName, mozIDownloadMonitor* aDownloadMonitor,
			       CmguiICommandData* aCommandData)
  : mDownloadMonitor(aDownloadMonitor), mMemoryName(aMemoryName),
    mCommandData(aCommandData)
{
	mMemory = NULL;
}

DnloadListener::~DnloadListener()
{
	if (mMemory)
	{
		PR_Free(mMemory);
	}
}


NS_IMPL_ISUPPORTS2(DnloadListener,
                   nsIRequestObserver,
                   nsIStreamListener);

NS_IMETHODIMP
DnloadListener::OnStartRequest(nsIRequest* request, nsISupports* context)
{
  //printf( "DnloadListener::OnStartRequest entered\n");

  mReadChunk=10000;

  mMemory = PR_Malloc(mReadChunk*sizeof(char));
  if (!mMemory)
  {
    printf( "%s %d: Malloc error\n",(char*)__FILE__, (int)__LINE__);
  }
  mMemoryLength = mReadChunk;
  mTotalRead=0;

  //printf( "DnloadListener::OnStartRequest exited\n");
  return NS_OK;
}



NS_IMETHODIMP
DnloadListener::OnDataAvailable(nsIRequest* request,
                              nsISupports* context,
                              nsIInputStream *aStream, 
                              PRUint32 offset,
                              PRUint32 aLength)
{
  int return_code;
  PRUint32 bytesRead;
  void  *new_data;
  char *stream_start;
  nsresult rv = NS_OK;

  //printf("DnloadListener::OnDataAvailable [offset=%u length=%u]\n",offset, aLength);
  
  return_code = 1;
  while ( return_code ) 
  {
    if (mTotalRead + mReadChunk > mMemoryLength)
    {
      if ((new_data = (char *)PR_Realloc((void *)mMemory, (mMemoryLength + mReadChunk))))
      {
        mMemory = new_data;
        mMemoryLength += mReadChunk;
      }
      else
      {
        printf("%s %d: Realloc error, rv=0x%08X\n",(char*)__FILE__,
               (int)__LINE__, (int)rv);
        return_code = 0;
      }
    }
    if (return_code)
    {
      stream_start = (char *)mMemory + mTotalRead;
      rv = aStream->Read(stream_start, mReadChunk, &bytesRead );
      if ( NS_SUCCEEDED( rv ) ) 
      {
        mTotalRead += bytesRead;
        if ( bytesRead < mReadChunk)
	{
          return_code = 0;
	} 
      } 
      else 
      {
        printf( "%s %d: Read error, rv=0x%08X\n",(char*)__FILE__, (int)__LINE__, (int)rv );
        return_code = 0;
      }
    }
  }

  if (mMemoryLength != mTotalRead)
  {
    if ((new_data = PR_Realloc(mMemory, mTotalRead*sizeof(char))))
      {
        mMemory = new_data;
        mMemoryLength = mTotalRead;
      }
      else
      {
        printf( "%s %d: Realloc error, rv=0x%08X\n",(char*)__FILE__, (int)__LINE__, (int)rv );
        return_code = 0;
      }
  }

  //printf( "DnloadListener::OnDataAvailable exit \n");
  return NS_OK;
}


NS_IMETHODIMP
DnloadListener::OnStopRequest(nsIRequest* request, nsISupports* context,
                            nsresult aStatus)
{
  //printf( "DnloadListener::OnStopRequest\n");
  
  //puts((char *)mMemory);
  char *name = ToNewCString(mMemoryName);
  mCommandData->DefineMemoryBlock(name, NS_REINTERPRET_CAST(PRUint8*,mMemory), mMemoryLength);
  if (mDownloadMonitor != nsnull)
    mDownloadMonitor->DecrementCounter();

  mCommandData = nsnull;
  mDownloadMonitor = nsnull;

  return NS_OK;
}



NS_IMETHODIMP
mozCmgui::UiSafeRun(nsIRunnable* aRunnable)
{
#ifndef XP_WIN
  return aRunnable->Run();
#else
  if (corbaThread != nsnull)
  {
    return corbaThread->RunOnThread(aRunnable);
  }
  else
  {
    printf("corba thread is null!\n");
    fflush(stdout);
	 return NS_ERROR_ABORT;
  }
#endif
}

#ifdef XP_WIN32
NS_IMPL_ISUPPORTS1(mozCmguiCORBAThread, nsIRunnable);

mozCmguiCORBAThread::mozCmguiCORBAThread()
  : mAutoDestruct(false)
{
  mCORBASendSemaphore = CreateSemaphore(NULL, 0, 1, NULL);
  mCORBARecvSemaphore = CreateSemaphore(NULL, 0, 1, NULL);

  // While both semaphores are down, launch the thread...
  nsCOMPtr<nsIThread> thread;
  NS_ADDREF(this);
  NS_NewThread(getter_AddRefs(thread), this);
}

mozCmguiCORBAThread::~mozCmguiCORBAThread()
{
  // This is only supposed to be called *after* the thread has noticed that
  // mAutoDestruct is on...
  CloseHandle(mCORBASendSemaphore);
  CloseHandle(mCORBARecvSemaphore);
}

NS_IMETHODIMP
mozCmguiCORBAThread::Run()
{
  while (true)
  {
    // Wait for the send semaphore to go up...
    WaitForSingleObject(mCORBASendSemaphore, INFINITE);

    // If we are to be destroyed, do it now...
    if (mAutoDestruct)
    {
      mozCmguiCORBAThread* ct = this;
      NS_RELEASE(ct);
      return NS_OK;
    }
   
    // Do our work...
    mReturnCode = mDoRun->Run();

    // Signal back that we are done...
    ReleaseSemaphore(mCORBARecvSemaphore, 1, NULL);
  }
}

NS_IMETHODIMP
mozCmguiCORBAThread::RunOnThread(nsIRunnable* aDoRun)
{

  assert(aDoRun != NULL);

  // We can safely touch this until we signal mCORBASendSemaphore(assuming we
  // were called on the main thread).
  mDoRun = aDoRun;
  
  // Raise the send semaphore...
  ReleaseSemaphore(mCORBASendSemaphore, 1, NULL);

  // Wait for the recv semaphore to go up, and process windows messages at
  // the same time...
  while
  (
   MsgWaitForMultipleObjects(1, &mCORBARecvSemaphore, FALSE, INFINITE, QS_ALLEVENTS) ==
   WAIT_OBJECT_0 + 1
  )
  {
    MSG msg;
    GetMessage(&msg, NULL, 0, 0);
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }

  return (mReturnCode);
}

void
mozCmguiCORBAThread::AutoDestruct()
{
  // We can safely touch this until we signal mCORBASendSemaphore(assuming we
  // were called on the main thread).
  mAutoDestruct = true;
  
  // Raise the send semaphore...
  ReleaseSemaphore(mCORBASendSemaphore, 1, NULL);
}

#endif
